import functools
import re

version_patt = re.compile(r"^(\d+)\.(\d+)\.(\d+)(-(.+))?")


@functools.total_ordering
class Version:
    def __init__(self, string: str):
        m = version_patt.match(string)
        if not m:
            raise ValueError(f"{string} is not a valid version")
        self._x = int(m.group(1))
        self._y = int(m.group(2))
        self._z = int(m.group(3))
        self._suffix = m.group(5)

    def __repr__(self):
        v = f"{self._x}.{self._y}.{self._z}"
        return v if not self._suffix else f"{v}-{self._suffix}"

    def __eq__(self, other):
        return self._x == other._x \
            and self._y == other._y \
            and self._z == other._z \
            and self._suffix == other._suffix

    def __lt__(self, other):
        """
        >>> Version('1.1.1') < Version('1.1.2')
        True
        >>> Version('1.1.1') < Version('1.2.1')
        True
        >>> Version('1.1.1') < Version('2.1.1')
        True
        >>> Version('1.1.1') < Version('1.1.1')
        False
        >>> strings = ['10.0.0', '10.0.1', '1.10.10', '1.1.1-SNAPSHOT', '9.9.9', '1.1.1']
        >>> vs = [Version(s) for s in strings]
        >>> max(vs)
        10.0.1
        >>> min(vs)
        1.1.1-SNAPSHOT
        """
        if self._x < other._x: return True
        if self._x > other._x: return False

        if self._y < other._y: return True
        if self._y > other._y: return False

        if self._z < other._z: return True
        if self._z > other._z: return False

        if not self._suffix and other._suffix: return False
        if self._suffix and not other._suffix: return True
        if not self._suffix and not other._suffix: return False
        return self._suffix < other._suffix

    def is_snapshot(self):
        """
        >>> Version("1.1.1").is_snapshot()
        False
        >>> Version("1.1.1-SNAPSHOT").is_snapshot()
        True
        """
        return self._suffix is not None and "SNAPSHOT" in self._suffix

    @staticmethod
    def is_valid(string):
        """Check if a string is a valid version number

        >>> test="1.1.1"
        >>> missing_one = [test[:i] + test[i + 1:] for i in range(len(test))]
        >>> all(not Version.is_valid(tc) for tc in missing_one)
        True
        >>> valid_strings = ["1.2.3", "20.50.1", "1.1.000-ASDF"]
        >>> all(Version.is_valid(tc) for tc in valid_strings)
        True
        """
        return version_patt.match(string) is not None


if __name__ == "__main__":
    import doctest

    doctest.testmod()



import shutil
from pathlib import Path
from typing import List, Generator
from urllib.request import urlopen

import lxml.html

from .files import get_cache_destination, get_distro_name
from .utils import Version

_default_repo = "https://repo.icatproject.org/repo/org/icatproject"


def _extract_content_list_from_table(page: str) -> Generator[str, None, None]:
    parsed = lxml.html.fromstring(page)
    table_rows = parsed.xpath("body/table//tr")
    # First rows are headers/dividers, last is a divider
    content_rows = table_rows[3:-1]
    # Second column contains a link element, whose href is what we want
    links = (r.getchildren()[1].getchildren()[0] for r in content_rows)
    return (l.get("href").strip("/") for l in links)


def get_components(repo_url=_default_repo) -> List[str]:
    with urlopen(repo_url) as f:
        if f.status != 200:
            raise Exception(f"Couldn't reach {repo_url}")
        page = f.read().decode()
        return list(_extract_content_list_from_table(page))


def get_component_versions(
        component: str,
        repo_url=_default_repo) -> List[Version]:
    with urlopen(f"{repo_url}/{component}") as f:
        if f.status != 200:
            raise Exception(f"Couldn't reach {repo_url}")
        page = f.read().decode()
        version_strings = _extract_content_list_from_table(page)
        return [Version(v) for v in version_strings if Version.is_valid(v)]


def get_distros(
        component: str,
        version: Version,
        repo_url=_default_repo) -> List[str]:
    with urlopen(f"{repo_url}/{component}/{version}") as f:
        if f.status != 200:
            raise Exception(f"Couldn't reach {repo_url}")
        page = f.read().decode()
        return [f for f in _extract_content_list_from_table(page)
                if f.endswith("distro.zip")]


def download_distro(
        component: str,
        version: Version,
        repo_url=_default_repo) -> Path:

    fname = get_distro_name(component, version)
    if not version.is_snapshot():
        url = f"{repo_url}/{component}/{version}/{fname}"
        print(f"Getting {url}")
    else:
        fname = get_distros(component, version)[-1]
        url = f"{repo_url}/{component}/{version}/{fname}"

    dest = get_cache_destination(fname)

    with urlopen(url) as response, open(dest, "wb") as f:
        if response.status != 200:
            raise Exception(f"Couldn't reach {repo_url}")
        shutil.copyfileobj(response, f)
    return dest


if __name__ == "__main__":
    from pprint import pprint

    pprint(download_distro("icat.server", Version("6.1.0-SNAPSHOT")))



import shutil
from pathlib import Path
from typing import List, Dict
from zipfile import ZipFile

from .utils import Version

_default_install_location = Path.home() / "install"
_default_cache_location = Path.home() / ".ipm" / "cache"

Path.mkdir(_default_cache_location, parents=True, exist_ok=True)
Path.mkdir(_default_install_location, parents=True, exist_ok=True)


def get_distro_name(component: str, version: Version) -> str:
    return f"{component}-{version}-distro.zip"


def get_cache_destination(file):
    return _default_cache_location / file


def _get_install_dir(component: str, version: Version) -> Path:
    return _default_install_location / component / str(version)


def extract_distro(component: str, version: Version) -> Path:
    fname = get_distro_name(component, version)
    dest = _get_install_dir(component, version)
    with ZipFile(get_cache_destination(fname), "r") as zf:
        zf.extractall(_default_install_location / component)
        shutil.move(
            _default_install_location / component / component,
            dest)
    return dest


def get_installed_packages(
        d=_default_install_location) -> Dict[str, List[Version]]:
    installed_components = [c for c in d.iterdir() if c.is_dir()]
    return {
        c.name: [Version(v.name) for v in c.iterdir() if Version.is_valid(v.name)]
        for c in installed_components
    }


if __name__ == "__main__":
    print(get_installed_packages())


def copy_config(
        component: str,
        src: Version,
        dest: Version):
    """Copy .properties and .xml files from one installed package to another"""

    src_dir = _get_install_dir(component, src)
    dest_dir = _get_install_dir(component, dest)

    for f in src_dir.glob("*.properties"):
        shutil.copy(f, dest_dir)

    for f in src_dir.glob("*.xml"):
        shutil.copy(f, dest_dir)


if __name__ == "__main__":
    copy_config("icat.server", Version("5.0.1"), Version("6.0.0"))



from typing import Optional

from . import files
from . import repo
from .utils import Version


def _get_component_versions(component: str, installed_only: bool):
    installed_only_packages = files.get_installed_packages()
    installed_only_versions = installed_only_packages.get(component, [])

    if installed_only:
        for v in installed_only_versions:
            yield v
    else:
        available_versions = repo.get_component_versions(component)
        for v in available_versions:
            if v in installed_only_versions:
                yield f"{v} (installed)"
            else:
                yield v


def _get_packages(installed_only: bool):
    installed_only_packages = files.get_installed_packages()

    if installed_only:
        for package in installed_only_packages:
            current = max(installed_only_packages[package])
            yield f"{package} - {current}"
    else:
        available = repo.get_components()
        for package in available:
            if package in installed_only_packages:
                current = max(installed_only_packages[package])
                yield f"{package} (installed: {current})"
            else:
                yield package


def do_list(component: Optional[str], installed_only: bool):
    if component:
        source = _get_component_versions(component, installed_only)
    else:
        source = _get_packages(installed_only)

    for line in source:
        print(line)


def do_install(
        component: str,
        version: Optional[Version] = None,
        allow_snapshots=False):
    installed = files.get_installed_packages().get(component, [])

    if version:
        if version in installed:
            print(f"{component} {version} is already installed")
            return
        install_version = version
    else:
        all_available = repo.get_component_versions(component)
        available = [v for v in all_available if
                     (allow_snapshots or not v.is_snapshot())]
        latest = max(available)
        if latest in installed:
            print(f"Latest available version, {latest}, is already installed")
            if latest != max(all_available):
                print(f"Newer snapshot, {max(all_available)}, is available")
            return
        install_version = latest

    repo.download_distro(component, install_version)
    dest = files.extract_distro(component, install_version)

    if not installed:
        print(
            f"Installed {dest}. No prior version existed, so configuration "
            "must be completed manually")
        return

    print(f"Copying config from existing install to {dest}")
    files.copy_config(component, max(installed), install_version)


def do_upgrade(component: str, allow_snapshots: bool = False):
    installed_versions = files.get_installed_packages().get(component, [])
    if not installed_versions:
        print(f"No existing {component} install to upgrade")
        return
    do_install(component, allow_snapshots=allow_snapshots)



import argparse

from . import commands
from .utils import Version


def run():
    parser = argparse.ArgumentParser(
        description="Find and install ICAT components")
    subparsers = parser.add_subparsers()

    list_parser = subparsers.add_parser(
        "list",
        help="Show available/installed components")
    list_parser.add_argument(
        "component",
        nargs="?",
        help="List available versions of a specific component")
    list_parser.add_argument(
        "-i", "--installed",
        action="store_true",
        dest="installed_only",
        help="Show only installed components, or only installed versions for a "
             "components")
    list_parser.set_defaults(func=commands.do_list)

    install_parser = subparsers.add_parser(
        "install",
        help="Install an ICAT package"
    )
    install_parser.add_argument(
        "component",
        help="The component to install")
    install_parser.add_argument(
        "version",
        help="Specific version to install. Defaults to latest version if not "
             "specified",
        type=Version,
        nargs="?")
    install_parser.add_argument(
        "-s", "--allow-snapshots",
        action="store_true",
        help="Allow snapshot versions. If not set, only non -SNAPSHOT versions "
             "will be used"
    )
    install_parser.set_defaults(func=commands.do_install)

    upgrade_parser = subparsers.add_parser(
        "upgrade",
        help="Upgrade to the newest available version of a component"
    )
    upgrade_parser.add_argument(
        "component",
        help="The component to upgrade")
    upgrade_parser.add_argument(
        "-s", "--allow-snapshots",
        action="store_true",
        help="Allow snapshot versions. If not set, only non -SNAPSHOT versions "
             "will be used"
    )
    upgrade_parser.set_defaults(func=commands.do_upgrade)

    args = parser.parse_args()
    if not hasattr(args, 'func'):
        print("Must provide a subcommand")
        parser.print_usage()
        exit(1)

    kwargs = {
        k: v for k, v in vars(args).items() if k != "func"
    }
    args.func(**kwargs)


if __name__ == "__main__":
    run()
